# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""SQL Parser."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

import datetime
import fnmatch
import functools
import hashlib
import math
import operator
import random
from typing import Any, Sequence, Union
import uuid
import numpy as np
from flib.base import parser, phonetic
from flib.base.parser import Symbol, UNARY, BINARY, FUNCTION, CONSTANT
from pandb.core import dtype

#
# SQL Operators
#

class SQLOperators(parser.Vocabulary):
    """SQL:2016 Clause Operator Vocabulary.

    This Vocabulary follows the specifications of the ISO standard SQL:2016
    (ISO 9075:2016), which is developed by a common committee of ISO and IEC.

    """

    def __init__(self) -> None:
        super().__init__()

        def sql_and(a: Any, b: Any) -> bool:
            """SQL-AND Operator."""
            return a and b

        def sql_or(a: Any, b: Any) -> bool:
            """SQL-OR Operator."""
            return a or b

        def sql_in(a: Any, b: Sequence) -> bool:
            """SQL-IN Operator."""
            return a in b

        def sql_like(string: str, pattern: str) -> bool:
            """SQL-LIKE Operator.

            The LIKE operator is used in a WHERE clause to search for a
            specified pattern in a column. There are two wildcards used in
            conjunction with the LIKE operator: The percent (%) sign represents
            zero, one, or multiple characters. The underscore (_)represents a
            single character.

            """
            # Translate SQL wildcards to Unix wildcards
            # TODO: Better use regular expressions, since the translation
            # approach raises the problem, that '*' and '?' are not allowed
            # anymore
            pattern = pattern.translate(str.maketrans('%_', '*?'))

            # Use fnmatch.fnmatch to match the given string
            return fnmatch.fnmatch(string, pattern)

        # Binding Operators
        self.update([
            Symbol(BINARY, ',', parser._pack, 30, True), # pylint: disable=W0212
            Symbol(BINARY, '||', operator.concat, 1)])

        # Arithmetic Operators
        self.update([
            Symbol(UNARY, '+', operator.pos, 11, True), # Unary Plus
            Symbol(UNARY, '-', operator.neg, 11, True), # Negation
            Symbol(BINARY, '/', operator.truediv, 10, True), # Division
            Symbol(BINARY, '%', operator.mod, 10, True), # Remainder
            Symbol(BINARY, '*', operator.mul, 10, True), # Multiplication
            Symbol(BINARY, '+', operator.add, 9, True), # Addition
            Symbol(BINARY, '-', operator.sub, 9, True)]) # Subtraction

        # Bitwise Operators
        self.update([
            Symbol(BINARY, '&', operator.and_, 7, True), # Bitwise AND
            Symbol(BINARY, '^', operator.xor, 6, True), # Bitwise XOR
            Symbol(BINARY, '|', operator.or_, 5, True)]) # Bitwise OR

        # Comparison Operators
        self.update([
            Symbol(BINARY, '=', operator.eq, 4), # Equality
            Symbol(BINARY, '<>', operator.ne, 4), # Inequality
            Symbol(BINARY, '>', operator.gt, 4, True), # Greater
            Symbol(BINARY, '<', operator.lt, 4, True), # Lower
            Symbol(BINARY, '>=', operator.ge, 4, True), # Greater or Equal
            Symbol(BINARY, '<=', operator.le, 4, True), # Lower or Equal
            Symbol(BINARY, 'IN', sql_in, 4), # Containment
            Symbol(BINARY, 'LIKE', sql_like, 4)]) # Matching

        # Logical Operators
        # TODO: ALL, ANY, BETWEEN, EXISTS, SOME
        self.update([
            Symbol(UNARY, 'NOT', operator.not_, 3), # Boolean NOT
            Symbol(BINARY, 'AND', sql_and, 2), # Boolean AND
            Symbol(BINARY, 'OR', sql_or, 1)]) # Boolean OR

        # Compound Operators
        # Hint: For immutable targets such as strings, numbers, and tuples,
        # the updated value is computed, but not assigned back to the input
        # variable
        self.update([
            Symbol(BINARY, '+=', operator.iadd, 0),
            Symbol(BINARY, '-=', operator.isub, 0),
            Symbol(BINARY, '*=', operator.imul, 0),
            Symbol(BINARY, '/=', operator.itruediv, 0),
            Symbol(BINARY, '%=', operator.imod, 0),
            Symbol(BINARY, '&=', operator.iand, 0),
            Symbol(BINARY, '^-=', operator.ixor, 0),
            Symbol(BINARY, '|*=', operator.ior, 0)])

#
# SQL Functions
#

class SQLFunctions(SQLOperators):
    """SQL:2016 Clause Operator and Function Vocabulary.

    This Vocabulary follows the specifications of the ISO standard SQL:2016
    (ISO 9075:2016), which is developed by a common committee of ISO and IEC.

    """
    def __init__(self) -> None:
        super().__init__()

        def sql_covar_pop(seqa: Sequence, seqb: Sequence) -> Any:
            """SQL-COVAR_POP Function."""
            return np.cov(seqa, seqb, ddof=0)[0, 1]

        def sql_covar_samp(seqa: Sequence, seqb: Sequence) -> Any:
            """SQL-COVAR_SAMP Function."""
            return np.cov(seqa, seqb, ddof=1)[0, 1]

        def sql_locate(search: str, string: str, start: int = 0) -> int:
            """SQL-LOCATE Function."""
            return string.find(search, start)

        def sql_lpad(string: str, width: int, fillchar: str = ' ') -> str:
            """SQL-LPAD Function."""
            return string.rjust(width, fillchar)

        def sql_rpad(string: str, width: int, fillchar: str = ' ') -> str:
            """SQL-RPAD Function."""
            return string.ljust(width, fillchar)

        def sql_repeat(string: str, count: int) -> str:
            """SQL-REPEAT Function."""
            return string * count

        def sql_space(count: int) -> str:
            """SQL-SPACE Function."""
            return ' ' * count

        def sql_substr(string: str, start: int, length: int = 0) -> str:
            """SQL-SUBSTRING Function."""
            return string[start: start + length] if length else string[start:]

        def sql_translate(string: str, in_chars: str, out_chars: str) -> str:
            """SQL-TRANSLATE Function."""
            return string.translate(str.maketrans(in_chars, out_chars))

        def sql_octet_length(string: str) -> int:
            """SQL-OCTET_LENGTH Function."""
            return len(string.encode('utf-8'))

        def sql_greatest(*args: Any) -> Any:
            """SQL-GREATEST Function."""
            return max(args)

        def sql_least(*args: Any) -> Any:
            """SQL-LEAST Function."""
            return min(args)

        def sql_md5(string: str) -> str:
            """SQL-MD5 Function."""
            return hashlib.md5(string.encode('utf-8')).hexdigest()

        def sql_sha1(string: str) -> str:
            """SQL-SHA1 Function."""
            return hashlib.sha1(string.encode('utf-8')).hexdigest()

        def sql_cot(x: Union[float, int, bool]) -> float:
            """SQL-COT Function."""
            return 1 / math.tan(x)

        def sql_date() -> dtype.Date:
            """SQL-CURRENT_DATE Function."""
            return datetime.datetime.now().date()

        def sql_time() -> dtype.Time:
            """SQL-CURRENT_TIME Function."""
            return datetime.datetime.now().time()

        def sql_timestamp() -> dtype.TimeStamp:
            """SQL-CURRENT_TIMESTAMP Function."""
            return datetime.datetime.now()

        sql_stddev_pop = functools.partial(np.std, ddof=0)
        sql_stddev_samp = functools.partial(np.std, ddof=1)
        sql_var_pop = functools.partial(np.var, ddof=0)
        sql_var_samp = functools.partial(np.var, ddof=1)

        # # Function Argument Binding
        # self.update([
        #     Symbol(BINARY, 'AS', parser._pack, 30),
        #     Symbol(BINARY, '||', operator.concat, 1)])
        #     #Symbol(BINARY, 'AS', parser._pack, 13, True),

        # Aggregate Functions
        self.update([
            Symbol(FUNCTION, 'COUNT', len, 20),
            Symbol(FUNCTION, 'MIN', min, 20),
            Symbol(FUNCTION, 'MAX', max, 20),
            Symbol(FUNCTION, 'SUM', sum, 20),
            Symbol(FUNCTION, 'AVG', np.mean, 20),
            Symbol(FUNCTION, 'STDDEV_POP', sql_stddev_pop, 20),
            Symbol(FUNCTION, 'STDDEV_SAMP', sql_stddev_samp, 20),
            Symbol(FUNCTION, 'VAR_POP', sql_var_pop, 20),
            Symbol(FUNCTION, 'VAR_SAMP', sql_var_samp, 20),
            Symbol(FUNCTION, 'COVAR_POP', sql_covar_pop, 20),
            Symbol(FUNCTION, 'COVAR_SAMP', sql_covar_samp, 20)])

        # String Functions
        # TODO:
        # POSITION(search IN str) -> Requires RegEx Operator definition
        # QUOTE(x) Quote SQL in string x
        self.update([
            Symbol(FUNCTION, 'ASCII', ascii, 20),
            Symbol(FUNCTION, 'CHR', chr, 20),
            Symbol(FUNCTION, 'CONCAT', operator.concat, 20),
            Symbol(FUNCTION, 'LOCATE', sql_locate, 20),
            Symbol(FUNCTION, 'LOWER', str.lower, 20),
            Symbol(FUNCTION, 'UPPER', str.upper, 20),
            Symbol(FUNCTION, 'LPAD', sql_lpad, 20),
            Symbol(FUNCTION, 'RPAD', sql_rpad, 20),
            Symbol(FUNCTION, 'LTRIM', str.lstrip, 20),
            Symbol(FUNCTION, 'RTRIM', str.rstrip, 20),
            Symbol(FUNCTION, 'TRIM', str.strip, 20),
            Symbol(FUNCTION, 'REPEAT', sql_repeat, 20),
            Symbol(FUNCTION, 'SPACE', sql_space, 20),
            Symbol(FUNCTION, 'CHAR', str, 20),
            Symbol(FUNCTION, 'SUBSTR', sql_substr, 20),
            Symbol(FUNCTION, 'REPLACE', str.replace, 20),
            Symbol(FUNCTION, 'INITCAP', str.capitalize, 20),
            Symbol(FUNCTION, 'TRANSLATE', sql_translate, 20),
            Symbol(FUNCTION, 'LENGTH', len, 20),
            Symbol(FUNCTION, 'OCTET_LENGTH', sql_octet_length, 20),
            Symbol(FUNCTION, 'GREATEST', sql_greatest, 20),
            Symbol(FUNCTION, 'LEAST', sql_least, 20),
            Symbol(FUNCTION, 'SOUNDEX', phonetic.soundex, 20),
            Symbol(FUNCTION, 'MD5', sql_md5, 20),
            Symbol(FUNCTION, 'SHA1', sql_sha1, 20),
            Symbol(FUNCTION, 'UUID', uuid.uuid1, 20)])

        # Trigonometric Functions
        self.update([
            Symbol(FUNCTION, 'ASIN', math.asin, 20),
            Symbol(FUNCTION, 'ACOS', math.acos, 20),
            Symbol(FUNCTION, 'ATAN', math.atan, 20),
            Symbol(FUNCTION, 'ATAN2', math.atan2, 20),
            Symbol(FUNCTION, 'SIN', math.sin, 20),
            Symbol(FUNCTION, 'COS', math.cos, 20),
            Symbol(FUNCTION, 'TAN', math.tan, 20),
            Symbol(FUNCTION, 'COT', sql_cot, 20),
            Symbol(FUNCTION, 'SINH', math.sinh, 20),
            Symbol(FUNCTION, 'COSH', math.cosh, 20),
            Symbol(FUNCTION, 'TANH', math.tanh, 20),
            Symbol(FUNCTION, 'ATANH', math.atanh, 20)])

        # Numeric Functions
        self.update([
            Symbol(FUNCTION, 'ABS', abs, 20),
            Symbol(FUNCTION, 'SIGN', np.sign, 20),
            Symbol(FUNCTION, 'MOD', operator.mod, 20),
            Symbol(FUNCTION, 'CEIL', math.ceil, 20),
            Symbol(FUNCTION, 'FLOOR', math.floor, 20),
            Symbol(FUNCTION, 'ROUND', round, 20),
            Symbol(FUNCTION, 'TRUNCATE', math.trunc, 20),
            Symbol(FUNCTION, 'SQRT', math.sqrt, 20),
            Symbol(FUNCTION, 'EXP', math.exp, 20),
            Symbol(FUNCTION, 'POWER', pow, 20),
            Symbol(FUNCTION, 'LN', math.log, 20),
            Symbol(FUNCTION, 'LOG', math.log, 20),
            Symbol(FUNCTION, 'LOG10', math.log10, 20),
            Symbol(FUNCTION, 'SETSEED', random.seed, 20),
            Symbol(FUNCTION, 'RAND', random.random, 20)])

        # Date and time functions
        # Allow Syntax: EXTRACT(YEAR FROM x)
        # -> add binding operator 'FROM'
        # -> add binding operator 'AS'
        self.update([
            Symbol(CONSTANT, 'CURRENT_DATE', sql_date, factory=True),
            Symbol(CONSTANT, 'CURRENT_TIME', sql_time, factory=True),
            Symbol(CONSTANT, 'CURRENT_TIMESTAMP', sql_timestamp, factory=True),
            # Symbol(FUNCTION, 'EXTRACT', abs, 20),
            # EXTRACT(MONTH FROM x)
            # EXTRACT(DAY FROM x)
            # EXTRACT(HOUR FROM x)
            ])

#
# Constructors
#

#def parse_projection(): ...

def parse_clause(
        clause: str, variables: parser.OptVars = None) -> parser.Expression:
    return parser.Parser(SQLFunctions()).parse(clause, variables=variables)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Database Tables."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

import types
from typing import Any, NewType, Tuple, List, Optional, Iterator, Union
from flib.base import abc, attrib, check, stype
from flib.errors import RowLookupError, ProxyError, InvalidTypeError
from flib.base.types import StrList, StrTuple, OptOp, SeqOp, OptType
from flib.base.types import OptStrTuple, OptInt, OptStr, Mapping, OptMapping
from pandb.core import record, cursor

#
# Structural Types
#

# Various
OrderByType = Optional[Union[str, StrList, StrTuple, SeqOp]]

# MappingProxy
MappingProxy = types.MappingProxyType
OptMappingProxy = Optional[MappingProxy]

# Fields
OptFieldTuple = Optional[Tuple[record.Field, ...]]

# Colums
SelColA = str # Select Column: name
SelColB = Tuple[str, SeqOp] # Select Column: name, aggregator
SelCol = Union[SelColA, SelColB] # Select Column
ColsDef = Tuple[record.ColDef, ...]
OptColsDef = Optional[ColsDef]

# Rows
Row = NewType('Row', record.Record)
OptRow = Optional[Row]
RowList = List[Row]
RowLike = Union[tuple, Mapping, Row]
RowLikeClasses = (tuple, Mapping, record.Record)
RowLikeList = Union[RowList, List[tuple], List[Mapping]]
ValuesType = Optional[Union[RowLike, RowLikeList]]

#
# Constants
#

PROXY_MODE_FLAG_CACHE = 0b0001
PROXY_MODE_FLAG_INCREMENTAL = 0b0010
PROXY_MODE_FLAG_READONLY = 0b0100

#
# Table Class
#

class Table(attrib.Group):
    """Table Class.

    Args:
        name: Optional table name. If provided, the choice of the table name is
            restricted to valid identifiers, described by [UAX31]_.
        columns: Optionl tuple of *column definitions*. All column definitions
            independent from each other can be given in one of the following
            formats: (1) In order to only specify the name of the column,
            without further information, the colum definition has to be given as
            a string `<name>`. Thereby the choice of `<name>` is restricted to
            valid identifiers, described by [UAX31]_. (2) If, additionally to
            the name, also the data type of the column shall be specified, then
            the column definition has to be given as a tuple `(<name>, <type>)`.
            Thereby `<type>` is required to by a valid :class:`type`, like
            like :class:`str`, :class:`int`, :class:`float` or :class:`Date
            <datetime.datetime>`. (3) Finally the column definition may also
            contain supplementary constraints and metadata. In this case the
            definition has to be given as a tuple `(<name>, <type>, <dict>)`,
            where `<dict>` is dictionary which comprises any items, documented
            by the function :func:`dataclasses.fields`.
        metadata: Optional dictionary, with supplementary metadata of the table.
            This does not comprise metadata of the fields, which has to be
            included within the field declarations.
        parent: Reference to parent :class:`attribute group
            <flib.base.attrib.Group>`, which is used for inheritance and
            shared attributes. By default no parent is referenced.

    """

    #
    # Public Attributes
    #

    name: property = attrib.Virtual(fget='_get_name', fset='_set_name')
    name.__doc__ = "Name of the table."

    metadata: property = attrib.Virtual(fget='_get_metadata_proxy')
    metadata.__doc__ = """
    Read-only attribute, that provides an access to the tables metadata by a
    :class:`MappingProxy <types.MappingProxyType>`. Individual entries can be
    accessed and changed by the methods :meth:`.get_metadata` and
    :meth:`.set_metadata`. This attribute is not used by the :class:`Table class
    <pandb.core.table.Table>` itself, but intended for data integration by
    third-party extensions.
    """

    fields: property = attrib.Virtual(fget='_get_fields')
    fields.__doc__ = """
    Read-only attribute, that provides information about the fields of the
    table, as returned by the function :func:`dataclasses.fields`.
    """

    columns: property = attrib.Virtual(fget='_get_columns')
    columns.__doc__ = """
    Read-only attribute containing a tuple with all column names of the table.
    The order of the column names reflects the order of the corresponding fields
    in the table.
    """

    #
    # Protected Attributes
    #

    _data: property = attrib.Content(dtype=list)
    _name: property = attrib.MetaData(dtype=str)
    _metadata: property = attrib.MetaData(dtype=Mapping)
    _metadata_proxy: property = attrib.Temporary(dtype=MappingProxy)
    _record: property = attrib.Temporary()
    _diff: property = attrib.Temporary(dtype=list, default=[])
    _index: property = attrib.Temporary(dtype=list, default=[])
    _iter_index: property = attrib.Temporary()

    #
    # Special Methods
    #

    def __init__(
            self, name: OptStr = None, columns: OptColsDef = None,
            metadata: OptMapping = None,
            parent: Optional[attrib.Group] = None) -> None:
        super().__init__(parent=parent) # Initialize Attribute Group

        # Initialize Table Structure
        if columns:
            self.create(name, columns, metadata=metadata)

    def __iter__(self) -> Iterator:
        self._iter_index = iter(self._index)
        return self

    def __next__(self) -> Row:
        row = self.row(next(self._iter_index))
        while not row:
            row = self.row(next(self._iter_index))
        return row

    def __len__(self) -> int:
        return len(self._index)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name='{self.name}')"

    #
    # Public Methods
    #

    def create(
            self, name: OptStr, columns: ColsDef,
            metadata: OptMapping = None) -> None:
        """Create table structure.

        This method is motivated by the The SQL `CREATE TABLE`_ statement and
        used to define and initialize the structure of the table and it's
        fields. This includes naming the table, defining and initializing the
        field parameters by column names, types and further constraints, like
        default values, and supplementary table metadata.

        Args:
            name: The table name is required to be a valid identifier as defined
                in [UAX31]_.
            columns: Tuple of *column definitions*. All column definitions
                independent from each other can be given in one of the following
                formats: (1) In order to only specify the name of the column,
                without further information, the colum definition has to be
                given as a string `<name>`. Thereby the choice of `<name>` is
                restricted to valid identifiers, described by [UAX31]_. (2) If,
                additionally to the name, also the data type of the column shall
                be specified, then the column definition has to be given as a
                tuple `(<name>, <type>)`. Thereby `<type>` is required to by a
                valid :class:`type`, like like :class:`str`, :class:`int`,
                :class:`float` or :class:`Date <datetime.datetime>`. (3) Finally
                the column definition may also contain supplementary constraints
                and metadata. In this case the definition has to be given as a
                tuple `(<name>, <type>, <dict>)`, where `<dict>` is dictionary
                which comprises any items, documented by the function
                :func:`dataclasses.fields`.
            metadata: Optional dictionary (or arbitrary mapping), with
                supplementary metadata of the table. This does not comprise
                metadata of the columns, which has to be included within the
                column definitions.

        .. _CREATE TABLE: https://en.wikipedia.org/wiki/Create_(SQL)

        """
        self._set_name(name) # Set Name of the Table
        self._create_metadata(metadata) # Set supplementary Metadata of Table
        self._build_record_class(*columns) # Build Record Class

    def drop(self) -> None:
        """Delete table data and table structure.

        This method is motivated by the the SQL `DROP TABLE`_ statement and used
        to delete the table data and the table structure, given by the field
        declarations, the table metadata and the table identifier.

        Warning:
            This operation should be treated with caution as it can not be
            reverted by calling :meth:`.rollback`.

        .. _DROP TABLE: https://en.wikipedia.org/wiki/Drop_(SQL)

        """
        self.truncate() # Delete Table Data
        self._delete_header() # Delete Table Structure
        self._delete_metadata() # Delete Table Metadata
        self._delete_name() # Delete Table Identifier

    def truncate(self) -> None:
        """Delete table data.

        This method is motivated by the SQL `TRUNCATE TABLE`_ statement and used
        to delete the data inside a table, but not the table structure and
        metadata.

        Warning:
            This operation should be treated with caution as it can not be
            reverted by calling :meth:`.rollback`.

        .. _TRUNCATE TABLE: https://en.wikipedia.org/wiki/Truncate_(SQL)

        """
        self._data = [] # Initialize Storage Table
        self._diff = [] # Initialize Diff Table
        self._index = [] # Initialize Table Master Index

    def commit(self) -> None:
        """Apply data changes to table.

        This method is motivated by the SQL `COMMIT`_ statement and applies
        all data :meth:`updates <.update>`, :meth:`inserts <.insert>` and
        :meth:`deletions <.delete>` since the creation of the table or
        the last :meth:`.commit` and makes all changes visible to other users.

        .. _COMMIT: https://en.wikipedia.org/wiki/Commit_(SQL)

        """
        # Update data table and index
        for rowid in list(range(len(self._data))):
            row = self.row(rowid)
            if not row:
                continue
            state = row._state # pylint: disable=W0212
            if state & record.STATE_DELETE:
                self._data[rowid] = None
                try:
                    self._index.remove(rowid)
                except ValueError:
                    pass
            elif state & (record.STATE_CREATE | record.STATE_UPDATE):
                self._data[rowid] = self._diff[rowid]
                self._data[rowid]._state = 0 # pylint: disable=W0212

        self._diff = [None] * len(self._data) # Initialize Diff Table

    def rollback(self) -> None:
        """Revert data changes from table.

        This method is motivated by the SQL `ROLLBACK`_ statement and reverts
        all data :meth:`updates <.update>`, :meth:`inserts <.insert>` and
        :meth:`deletions <.delete>` since the creation of the table or
        the last :meth:`.commit`.

        .. _ROLLBACK: https://en.wikipedia.org/wiki/Rollback_(SQL)

        """
        # Remove newly created rows from index and reset states of already
        # existing rows
        for rowid in list(range(len(self._data))):
            row = self.row(rowid)
            if not row:
                continue
            state = row._state # pylint: disable=W0212
            if state & record.STATE_CREATE:
                try:
                    self._index.remove(rowid)
                except ValueError:
                    pass
            else:
                self._data[rowid]._state = 0 # pylint: disable=W0212

        self._diff = [None] * len(self._data) # Initialize Diff Table

    def insert(
            self, values: ValuesType = None,
            columns: OptStrTuple = None) -> None:
        """Append one or more records to the table.

        This method is motivated by the SQL `INSERT`_ statement and appends one
        ore more records to the table. The data changes can be organized in
        transactions by :meth:`.commit` and :meth:`.rollback`.

        Args:
            values: Single record, given as :term:`row like` data or multiple
                records given as list of row like data. If the records are given
                as tuples, the corresponding column names are determined from
                the argument *columns*.
            columns: Optional tuple of known column names. By default the
                columns are taken from the attribute :attr:`.columns`.

        .. _INSERT: https://en.wikipedia.org/wiki/Insert_(SQL)

        """
        values = values or tuple([]) # Get default value tuple
        if not isinstance(values, list):
            self._append_row(values, columns)
            return
        for row in values:
            self._append_row(row, columns)

    def update(self, where: OptOp = None, **kwds: Any) -> None:
        """Update values of one or more records from the table.

        This method is motivated by the SQL `UPDATE`_ statement and changes the
        values of all records in the table, that satisfy the `WHERE`_ clause
        given by the keyword argument 'where'. The data changes can be organized
        in transactions by :meth:`.commit` and :meth:`.rollback`.

        Args:
            where: Optional filter operator, which determines, if a row is
                included within the result set or not. By default all rows are
                included within the result set.
            **kwds: Items, which keys are valid column names of the table, and
                the values the new data, stored in the corresponding fields.

        .. _DELETE: https://en.wikipedia.org/wiki/Delete_(SQL)
        .. _WHERE: https://en.wikipedia.org/wiki/Where_(SQL)

        """
        cur = cursor.Cursor(getter=self.row, where=where, parent=self)
        for row in cur:
            row._update(**kwds) # pylint: disable=W0212

    def delete(self, where: OptOp = None) -> None:
        """Delete one or more records from the table.

        This method is motivated by the SQL `DELETE`_ statement and marks all
        records in the table as deleted, that satisfy the `WHERE`_ clause given
        by the keyword argument 'where'. The data changes can be organized in
        transactions by :meth:`.commit` and :meth:`.rollback`.

        Args:
            where: Optional filter operator, which determines, if a row is
                included within the result set or not. By default all rows are
                included within the result set.

        .. _DELETE: https://en.wikipedia.org/wiki/Delete_(SQL)
        .. _WHERE: https://en.wikipedia.org/wiki/Where_(SQL)

        """
        cur = cursor.Cursor(getter=self.row, where=where, parent=self)
        for row in cur:
            row._delete() # pylint: disable=W0212

    def select(
            self, *args: stype.VarLike, where: OptOp = None,
            groupby: cursor.GroupByType = None, dtype: OptType = None,
            orderby: OrderByType = None, reverse: bool = False,
            batchsize: OptInt = None, mode: OptStr = None) -> cursor.Cursor:
        """Get cursor on a specified result set of records from table.

        This method is motivated by the SQL `SELECT`_ statement and creates a
        :class:`Cursor class <pandb.core.table.Cursor>` instance with specified
        properties.

        Args:
            *args: Optional columns of the result set. If provided the
                columns individually can be given can be given in one of the
                following formats: (1) In order to provide the content of the
                column the column has to be given by a string `<name>`, which is
                required to be known to the underlying object (table, view,
                etc.). (2) In order to aggregate the result set by using
                `groupby`, any non-grouping column has to be given as a tuple
                `(<name>, <operator>)`. Thereby `<operator>` is required to by a
                sequence operator, like :func:`max`, :func:`min` or :func:`sum`
                etc. If no colums is given, the arguments `groupby` and `dtype`
                are ignored, and the result set is returned as raw records.
            where: Optional filter operator, which determines, if a row is
                included within the result set or not. By default all rows are
                included within the result set.
            orderby: Optional parameter, that determine(s) the order of the rows
                within the result set. If provided, the parameter may be given
                as a column name, a tuple of column names or a callable sorting
                function. By default the order is determined by the creation
                order of the rows.
            reverse: Boolean value, which determines if the sorting order of the
                rows is reversed. For the default value ``False`` the sorting
                order is ascending with respect to given column names in the
                orderby parameter, for ``True`` it is descending.
            dtype: Format of the :term:`row like` data, which is used to
                represent the returned values of the result set. By default
                the result set is returned as a list of tuples.
            batchsize: Integer, that specifies the default number of rows which
                is to be fetched by the method :meth:`Cursor.fetch
                <pandb.core.cursor.Cursor.fetch>`. It defaults to 1, meaning to
                fetch a single row at a time. Whether and which batchsize to use
                depends on the application and should be considered with care.
                The batchsize can also be adapted during the lifetime of the
                cursor, which allows dynamic performance optimization.
            mode: Named string identifier for the cursor :py:attr:`.mode`. The
                default cursor mode is 'forward-only indexed'. Note: After
                initializing the curser, it's mode can not be changed anymore.

        Returns:
            New instance of :class:`Cursor class <pandb.core.cursor.Cursor>` on
            on a specified result set from the table.

        .. _SELECT: https://en.wikipedia.org/wiki/Select_(SQL)

        """
        # Check arguments and set default values
        if args:
            check.is_subset(
                'columns', set(args), 'table columns', set(self.columns))
        else:
            args = self.columns

        # Create and return cursor
        return cursor.Cursor(
            *args, getter=self.row, where=where, orderby=orderby,
            reverse=reverse, groupby=groupby, batchsize=batchsize, dtype=dtype,
            mode=mode, parent=self)

    def get_metadata(self, key: str) -> Any:
        """Get single entry from table metadata.

        Args:
            key: Name of metadata entry

        Returns:
            Value of metadata entry.

        """
        return self._metadata[key]

    def set_metadata(self, key: str, val: Any) -> None:
        """Change metadata entry of table."""
        self._metadata[key] = val

    def row(self, rowid: int) -> OptRow:
        """Get single row by row ID."""
        try:
            return self._diff[rowid] or self._data[rowid]
        except IndexError as err:
            raise RowLookupError(rowid) from err

    def pack(self) -> None:
        """Remove empty records from data and rebuild table index."""
        # Commit pending changes
        self.commit()

        # Remove empty records
        self._data = list(filter(None.__ne__, self._data))

        # Rebuild table index
        self._index = list(range(len(self._data)))
        for rowid in self._index:
            self._data[rowid]._id = rowid # pylint: disable=W0212

        # Rebuild diff table
        self._diff = [None] * len(self._data)

    #
    # Protected Methods
    #

    def _append_row(self, row: RowLike, columns: OptStrTuple = None) -> None:
        if columns:
            if not isinstance(row, tuple):
                raise TypeError() # TODO
            rec = self._create_record(**dict(zip(columns, row)))
        else:
            rec = self._create_record(row)
        self._data.append(None)
        self._diff.append(rec)
        self._append_rowid(rec._id) # pylint: disable=W0212

    def _append_rowid(self, rowid: int) -> None:
        self._index.append(rowid)

    def _create_metadata(self, mapping: OptMapping = None) -> None:
        check.has_opt_type("'metadata'", mapping, Mapping)
        self._metadata = mapping or {}
        self._metadata_proxy = MappingProxy(self._metadata)

    def _delete_metadata(self) -> None:
        del self._metadata_proxy
        del self._metadata

    def _build_record_class(self, *args: record.ColDef) -> None:
        # Dynamically create a new record class
        self._record = record.build(*args,
            newid=self._get_new_rowid,
            delete=self._remove_rowid,
            restore=self._append_rowid,
            update=self._update_row_diff,
            revoke=self._remove_row_diff)
        self.truncate() # Initialize table data

    def _create_record(self, data: RowLike) -> Row:
        if isinstance(data, tuple):
            return self._record(*data)
        if isinstance(data, Mapping):
            return self._record(**data)
        if isinstance(data, record.Record):
            keys = self.columns
            vals = tuple(getattr(data, key, None) for key in keys)
            return self._record(**dict(zip(keys, vals)))
        raise InvalidTypeError("'data'", data, RowLikeClasses)

    def _delete_header(self) -> None:
        self.truncate() # Delete table data
        del self._record # Delete record constructor

    def _get_new_rowid(self) -> int:
        return len(self._data)

    def _get_last_row(self, column: str) -> OptRow:
        try:
            return self._data[-1]
        except IndexError:
            return None

    def _get_columns(self) -> StrTuple:
        return tuple(field.name for field in self.fields)

    def _get_fields(self) -> OptFieldTuple:
        if not hasattr(self, '_record') or not self._record:
            return None
        return record.get_fields(self._record)

    def _get_name(self) -> str:
        return self._name or self._default_name()

    def _default_name(self) -> str:
        return 'unkown'

    def _delete_name(self) -> None:
        del self._name

    def _set_name(self, name: OptStr) -> None:
        if isinstance(name, str):
            check.is_identifier(f"'name'", name)
            self._name = name
        else:
            self._name = self._default_name()

    def _get_metadata_proxy(self) -> OptMappingProxy:
        return getattr(self, '_metadata_proxy', None)

    def _remove_row_diff(self, rowid: int) -> None:
        self._diff[rowid] = None

    def _remove_rowid(self, rowid: int) -> None:
        self._index.remove(rowid)

    def _update_row_diff(self, rowid: int, **kwds: Any) -> None:
        row = self.row(rowid)
        if not row: # Row has already been deleted
            return
        new = record.create_from(row, **kwds)
        self._diff[rowid] = new

#
# Proxy Class
#

class Proxy(Table, abc.Proxy):
    """Abstract Base Class for Table Proxies.

    Args:
        proxy_mode: Optional Integer, that determines the operation mode of the
            proxy.
        parent: Reference to parent :class:`attribute group
            <flib.base.attrib.Group>`, which is used for inheritance and
            shared attributes. By default no parent is referenced.

    """

    _proxy_mode: property = attrib.MetaData(dtype=int, default=1)

    #
    # Special Methods
    #

    def __init__(
            self, proxy_mode: OptInt = None,
            parent: Optional[attrib.Group] = None) -> None:

        # Initialize Proxy Base Class
        abc.Proxy.__init__(self)

        # Initialize Empty Table
        Table.__init__(self, parent=parent)

        # Initialize Table Proxy Parameters
        if proxy_mode is None:
            proxy_mode = PROXY_MODE_FLAG_CACHE # Set default proxy mode
        self._proxy_mode = proxy_mode

    def _post_init(self) -> None:
        # Retrieve all rows from source if table is cached
        if self._proxy_mode & PROXY_MODE_FLAG_CACHE:
            self.pull()

    #
    # Public Methods
    #

    def commit(self) -> None:
        """Push changes to source table and apply changes to local table."""
        if self._proxy_mode & PROXY_MODE_FLAG_READONLY:
            raise ProxyError('changes can not be commited in readonly mode')

        # For incremental updates of the source, the push request requires, that
        # changes have not yet been applied to the local table
        if self._proxy_mode & PROXY_MODE_FLAG_INCREMENTAL:
            self.push()
            super().commit()
            return
        # For full updates of the source, the push request requires, that all
        # changes have been applied to the local table
        super().commit()
        self.push()



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Table Proxy Classes."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

from pandb.core import table

#
# Classes
#

class Table(table.Proxy):
    """SQL-Table Proxy."""



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Cursor Class."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

import random
from typing import List, NewType, Tuple, Union, Optional
from flib.base import attrib, operator, stype
from flib.errors import InvalidTypeError, UserError
from flib.base.types import StrList, StrTuple, OptIntList, OptOp, Callable, OptInt
from flib.base.types import OptStr, Iterator, Mapping, SeqOp, AnyOp, OptType, BoolOp
from pandb.core import record

#
# Exceptions
#

class CursorError(UserError):
    """Base Class for Cursor Errors."""

class ModeError(CursorError):
    """Raise when a procedure is not supported by a cursor mode."""

    def __init__(self, mode: str, proc: OptStr = None) -> None:
        if proc:
            msg = f"{proc} is not supported by {mode} cursors"
        else:
            msg = f"unknown cursor mode '{mode}'"
        super().__init__(msg)

#
# Structural Types
#

# Various
PredLike = Optional[Union[str, BoolOp]]
OrderByType = Optional[Union[str, StrList, StrTuple, SeqOp]]
GroupByType = OrderByType
AggAttr = Union[str, Tuple[str, AnyOp]]
OptSeqOp = Optional[SeqOp]

# Variables
ColNames = Tuple[str, ...]
OptColNames = Optional[ColNames]

# Rows
Row = NewType('Row', record.Record)
OptRow = Optional[Row]
RowList = List[Row]
RowLike = Union[tuple, Mapping, Row]
RowLikeList = Union[RowList, List[tuple], List[Mapping]]

#
# Constants
#

MODE_BUFFERED = 0b0001
MODE_INDEXED = 0b0010
MODE_SCROLLABLE = 0b0100
MODE_RANDOM = 0b1000

#
# Cursor Class
#

class Cursor(attrib.Group):
    """Cursor Class.

    Args:
        *args: Optional :term:`variable defintions<variable defintion>`. If
            no variables are defined, the arguments `groupby` and `dtype` are
            ignored, and the result set is returned as raw records.
        where: Optional filter, which determines if a row is included within the
            result set or not. If provided, the filter ether can be given as
            a clause (string) which uses the field identifiers of the records as
            variables, or as a predicate (callable), which operates on the
            records. By default all rows are included within the result set.
        groupby:
        having:
        orderby: Optional parameter, that determine(s) the order of the rows
            within the result set. If provided, the parameter may be given
            as a column name, a tuple of column names or a callable sorting
            function. By default the order is determined by the creation
            order of the rows.
        index: List of row IDs, that are traversed by the cursor. By default the
            attribute '_index' of the parent object is used.
        mode: Named string identifier for the cursor :py:attr:`.mode`. The
            default cursor mode is 'forward-only indexed'. Note: After
            initializing the curser, it's mode can not be changed anymore.
        batchsize: Integer, that specifies the default number of rows which is
            to be fetched by the method :meth:`.fetch`. It defaults to 1,
            meaning to fetch a single row at a time. Whether and which batchsize
            to use depends on the application and should be considered with
            care. The batchsize can also be adapted during the lifetime of the
            cursor, which allows dynamic performance optimization.
        getter: Method, which is used to fetch single rows by their row ID.
        reverse: Boolean value, which determines if the sorting order of the
            rows is reversed. For the default value ``False`` the sorting
            order is ascending with respect to given column names in the
            orderby parameter, for ``True`` it is descending.
        parent: Reference to parent :class:'attribute group
            <flib.base.attrib.Group>', which is used for inheritance and
            shared attributes. By default no parent is referenced.
        dtype: Optional type of the returned records. Supported types are
            :class:`tuple` and :class:`dict`. The default type of the returned
            records depends on the variable definitions. If no variables are
            defined, the records by default are returned as instances of the
            :class:`Record class <pandb.core.record.Record>`, if variables are
            defined, then the records by default are returned as tuples.

    """

    #
    # Public Attributes
    #

    mode: property = attrib.Virtual('_get_mode')
    mode.__doc__ = """
    The read-only attribute :term:`cursor mode` provides information about the
    *scrolling type* and the *operation mode* of the cursor.
    """

    batchsize: property = attrib.MetaData(dtype=int, default=1)
    """
    The read-writable integer attribute *batchsize* specifies the default number
    of rows which is to be fetched by the method :meth:`.fetch`. It defaults
    to 1, meaning to fetch a single row at a time. Whether and which batchsize
    to use depends on the application and should be considered with care. The
    batchsize can also be adapted during the lifetime of the cursor, which
    allows dynamic performance optimization.
    """

    rowcount: property = attrib.Virtual('_get_rowcount')
    """
    The read-only integer attribute *rowcount* identifies the current number of
    rows within the cursor.
    """

    #
    # Protected Attributes
    #

    _mode_id: property = attrib.MetaData(dtype=int, factory='_default_mode')
    _index: property = attrib.MetaData(dtype=list, inherit=True)
    _getter: property = attrib.Temporary(dtype=Callable)
    _groupby: property = attrib.Temporary(dtype=Callable, default=None)
    _sorter: property = attrib.Temporary(dtype=Callable, default=None)
    _mapper: property = attrib.Temporary(dtype=Callable, default=None)
    _where: property = attrib.Temporary(dtype=Callable, default=None)
    _buffer: property = attrib.Temporary(dtype=list, default=[])

    #
    # Special Methods
    #

    def __init__(
            self, *args: stype.VarLike, where: PredLike = None,
            groupby: GroupByType = None, having: PredLike = None,
            orderby: OrderByType = None, reverse: bool = False,
            batchsize: OptInt = None, dtype: OptType = None,
            index: OptIntList = None, getter: OptOp = None,
            mode: OptStr = None, parent: Optional[attrib.Group] = None) -> None:
        # Initialize Attribute Group with parent Attribute Group
        super().__init__(parent=parent)

        # Update Cursor Parameters (The order is important)
        self._set_mode(mode)
        self._getter = getter
        self._set_filter(where)
        self._set_aggregator(*args, groupby=groupby, having=having)
        self._set_sorter(orderby, reverse=reverse)
        self._set_mapper(*args, dtype=dtype)

        # Initialize index
        if index is not None:
            self._index = index

        # Set mode and batchsize
        if batchsize:
            self.batchsize = batchsize

        # Initialize cursor
        if self._mode_id & MODE_INDEXED:
            self._create_index() # Initialize index
        if self._mode_id & MODE_BUFFERED:
            self._create_buffer() # Initialize buffer
        self.reset() # Initialize iterator

    def __iter__(self) -> Iterator:
        self.reset()
        return self

    def __next__(self) -> Row:
        return self.next()

    def __len__(self) -> int:
        return self.rowcount

    #
    # Public Methods
    #

    def fetch(self, size: OptInt = None) -> RowList:
        """Fetch rows from the result set.

        Args:
            size: Integer value, which represents the number of rows, which is
                fetched from the result set. For the given size -1 all remaining
                rows from the result set are fetched. By default the number of
                rows is given by the cursors attribute :attr:`.batchsize`.

        Returns:
            Result set given by a list of :term:`row like` data.

        """
        # TODO: Scrollable cursors are defined on sequences not on iterables:
        # the cursor can use operations, such as FIRST, LAST, PRIOR, NEXT,
        # RELATIVE n, ABSOLUTE n to navigate the results
        if size is None:
            size = self.batchsize
        if self._mode_id & MODE_RANDOM and size < 1:
            raise ModeError(self.mode, 'fetching all rows')
        finished = False
        rows: RowList = []
        while not finished:
            try:
                rows.append(self.next())
            except StopIteration:
                finished = True
            else:
                finished = 0 < size <= len(rows)

        if self._mapper:
            return list(map(self._mapper, rows))
        return rows

    def next(self) -> Row:
        """Return next row that matches the given filter."""
        mode = self._mode_id
        if mode & MODE_BUFFERED:
            return self._get_next_from_buffer()
        if mode & MODE_INDEXED:
            return self._get_next_from_fixed_index()
        # TODO: For dynamic cursors implement _get_next_from_dynamic_index()
        return self._get_next_from_fixed_index()

    def reset(self) -> None:
        """Reset cursor position before the first record."""
        mode = self._mode_id
        if mode & MODE_BUFFERED: # Iterate over fixed result set
            self._iter_buffer = iter(self._buffer)
        elif mode & MODE_INDEXED: # Iterate over fixed index
            self._iter_index = iter(self._index)
        else: # TODO: handle case for dynamic cursors by self._iter_table
            self._iter_index = iter(self._index)

    #
    # Protected Methods
    #

    def _create_index(self) -> None:
        if isinstance(self._index, list):
            self._index = self._index.copy()
        else:
            self._index = []

    def _create_buffer(self) -> None:
        # List all operators, which are required to create the result set in
        # the order of their appearance
        ops = [lambda seq: map(self._getter, seq)] # Getter
        if self._where:
            ops.append(lambda seq: filter(self._where, seq)) # Row filter
        if self._groupby:
            ops.append(self._groupby)
        if self._having:
            ops.append(lambda seq: filter(self._having, seq))
        if self._sorter:
            ops.append(self._sorter)
        ops.append(list)

        # Compose and apply the operators to the index
        self._buffer = operator.compose(*ops[::-1])(self._index)

    def _set_filter(self, where: PredLike = None) -> None:
        if where is None:
            self._where = None
            return

        if callable(where):
            # TODO: check if where is a valid record predicate
            self._where = where
            return

        if isinstance(where, str):
            self._where = operator.Lambda(where, domain=object)
            return

        raise InvalidTypeError('where', where, (type(None), str)) # TODO: Type!

    def _set_aggregator(
            self, *args: stype.VarLike, groupby: GroupByType = None,
            having: PredLike = None) -> None:

        if not groupby:
            self._groupby = None
            self._having = None
            return

        # In order to provide a grouped result set, the cursor requires to
        # preprocess the fields to variables. This in turn needs the cursor to
        # be buffered and countable in the sense to support counting the rows
        # (which is not possible for random cursors).
        if not args:
            raise CursorError(
                'group aggregation requires the definition of variables')
        if not self._mode_id & MODE_BUFFERED:
            raise ModeError(
                "group aggregation requires a static cursor "
                f"not '{self.mode}'")
        if self._mode_id & MODE_RANDOM:
            raise ModeError(
                "group aggregation requires a forward-only cursor "
                f"not '{self.mode}'")

        # Create group aggragation operator
        self._groupby = operator.create_group_aggregator(
            *args, key=groupby, domain=object)

        if having is None:
            self._having = None
            return

        if callable(having):
            # TODO: check if having is a valid group predicate
            self._having = having # type: ignore
            return

        # Get variable names
        variables = map(stype.create_variable, args)
        names = tuple(var.name for var in variables)

        if isinstance(having, str):
            self._having = operator.Lambda(having, domain=(tuple, names))
            return

        raise InvalidTypeError('having', having, (type(None), str)) # TODO

    def _set_sorter(
            self, orderby: OrderByType, reverse: bool = False) -> None:

        # If sorting is not used, the sorting operator is the identity
        if not (orderby or reverse):
            self._sorter = None
            return

        # Validate sorting parameters
        if self._mode_id & MODE_RANDOM:
            raise ModeError(
                f"sorting requires a finite cursor not '{self.mode}'")
        if not self._mode_id & MODE_BUFFERED:
            raise ModeError(
                f"sorting requires a static cursor not '{self.mode}'")

        # If 'orderby' is given as an operator, set it as sorter
        if callable(orderby):
            # TODO: check if orderby is a valid sorter
            self._sorter = orderby
            return

        # If 'orderby' is given as a string or as a tuple of strings, use it as
        # the sorting key of a created sorting operator
        if isinstance(orderby, str):
            keys = [orderby]
        elif isinstance(orderby, (list, tuple)):
            keys = list(orderby)
        else:
            keys = []
        domain = tuple if self._groupby else object
        self._sorter = operator.create_sorter(
            *keys, domain=domain, reverse=reverse)

    def _set_mapper(self, *args: stype.VarLike, dtype: OptType = None) -> None:
        # Validate Arguments
        if dtype and not args:
            raise CursorError(
                'mapping to a given target type '
                'requires the definition of variables')

        # Get variable names
        variables = map(stype.create_variable, args)
        names = tuple(var.name for var in variables)

        # If the result set is aggregated by a grouping operator, the mapper
        # acts on tuples as an itemgetter, which requires the specification of
        # the tuple variables.
        if self._groupby:
            self._mapper = operator.Getter(
                *names, domain=(tuple, names), target=dtype)
            return

        # If the result set is not aggregated, and the target type is specified,
        # then the mapper acts on record objects as an attribute getter with
        # a subsequent converter
        if dtype:
            self._mapper = operator.Getter(*names, domain=object, target=dtype)
            return

        # If the result set is not aggregated, and the target type is not
        # specified (or None), then the rows are returned as records and the
        # mapper is the identity function
        self._mapper = None

    def _default_mode(self) -> int:
        if self._sorter:
            return MODE_BUFFERED
        return MODE_INDEXED

    def _get_next_from_fixed_index(self) -> Row:
        is_random = self._mode_id & MODE_RANDOM
        matches = False
        while not matches:
            if is_random:
                rowid = random.randrange(len(self._index))
            else:
                rowid = next(self._iter_index)
            row = self._getter(rowid)
            if self._where:
                matches = self._where(row)
            else:
                matches = True
        return row

    def _get_next_from_buffer(self) -> Row:
        if self._mode_id & MODE_RANDOM:
            rowid = random.randrange(len(self._buffer))
            return self._buffer[rowid]
        return next(self._iter_buffer)

    def _get_mode(self) -> str:
        mode = self._mode_id
        tokens = []

        # Add name of traversal mode
        if mode & MODE_RANDOM:
            tokens.append('random')
        elif mode & MODE_SCROLLABLE:
            tokens.append('scrollable')

        # Add name of operation mode
        if mode & MODE_BUFFERED:
            tokens.append('static')
        elif mode & MODE_INDEXED:
            tokens.append('indexed')
        else:
            tokens.append('dynamic')

        return ' '.join(tokens)

    def _set_mode(self, name: OptStr = None) -> None:
        if name is None:
            return

        mode = 0
        name = name.strip(' ').lower()

        # Set traversal mode flags
        if 'random' in name:
            mode |= MODE_RANDOM
        elif 'scrollable' in name:
            mode |= MODE_SCROLLABLE

        # Set operation mode flags
        if 'static' in name:
            mode |= MODE_BUFFERED | MODE_INDEXED
        elif 'indexed' in name:
            mode |= MODE_INDEXED

        self._mode_id = mode

    def _get_rowcount(self) -> int:
        mode = self._mode_id
        if mode & MODE_RANDOM:
            raise ModeError(self.mode, 'counting rows')
        if mode & MODE_BUFFERED:
            return len(self._buffer)
        if self._where:
            raise ModeError(self.mode, 'counting filtered rows')
        return len(self._index)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Database and Data Warehouse."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

from flib.base import pkg
from pandb.core import table
from flib.errors import ConnectError
from flib.base.types import Any, Module

#
# Constructors
#

def connect(name: str, *args: Any, **kwds: Any) -> table.Proxy:
    """Connect Table Proxy

    Args:
        name: Name of module, which is used to connect a Table Proxy.
        *args: Arguments, that are passed to the class 'Table' of the given
            module.
        **kwds: Keyword arguments, that are passed to the class 'Table' of the
            given module.

    """
    module = pkg.get_submodule(name=name)
    if not isinstance(module, Module):
        raise ConnectError(f"module '{name}' does not exist")
    if not hasattr(module, 'Table'):
        raise ConnectError(f"module '{name}' does not contain a 'Table' class")
    return module.Table(*args, **kwds) # type: ignore



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Pandora database engine.

*Pandora* is a SQL database engine and a universal database proxy. Pandora aims
to provide a unified SQL:2016 interface to a broad variety of different data
sources, including SQL Databases and flat files from laboratory measurements.
Pandora is open source and based on the Python programming language.

"""

__version__ = '0.1.6'
__license__ = 'GPLv3'
__copyright__ = 'Copyright (c) 2019 Frootlab Developers'
__organization__ = 'frootlab'
__author__ = 'frootlab'
__email__ = 'frootlab@gmail.com'
__maintainer__ = 'Patrick Michl'
__docformat__ = 'google'
__authors__ = ['Patrick Michl <patrick.michl@gmail.com>']



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Record Class."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

import abc
import contextlib
import itertools
import dataclasses
from typing import Any, Union, Tuple, Type
from flib.base import check
from flib.errors import InvalidTypeError
from flib.base.types import StrDict, StrList, void, OptOp, TypeHint

#
# Structural Types
#

Field = dataclasses.Field
FieldTuple = Tuple[Field, ...]
ColDefA = str # Column definition: name
ColDefB = Tuple[str, type] # Column definition: name and type
ColDefC = Tuple[str, type, StrDict] # Column definition: name, type, constraints
ColDefD = Tuple[str, type, Field] # Column definition: name, type, constraints
ColDef = Union[ColDefA, ColDefB, ColDefC, ColDefD] # Column definition

#
# Constants
#

STATE_CREATE = 0b0001
STATE_UPDATE = 0b0010
STATE_DELETE = 0b0100

#
# Record Base Class
#

class Record(abc.ABC):
    """Abstract base class for :mod:`dataclasses` based records.

    Args:
        *args: Arguments, that are valid with respect to the column definitions
            of derived :mod:'dataclasses'.
        **kwds: Keyword arguments, that are valid with respect to the column
            definitions of derived :mod:'dataclasses'.

    """
    __slots__ = ['_id', '_name', '_state']

    _id: int
    _name: str
    _state: int

    @abc.abstractmethod
    def __init__(self, *args: Any) -> None:
        raise NotImplementedError()

    def __post_init__(self, *args: Any, **kwds: Any) -> None:
        self._validate()
        self._id = self._get_newid()
        self._state = STATE_CREATE

    def _validate(self) -> None:
        """Check validity of the field types."""
        fields = getattr(self, '__dataclass_fields__', {})
        for name, field in fields.items():
            if isinstance(field.type, str):
                continue # Do not type check structural types like 'typing.Any'
            value = getattr(self, name)
            check.has_type(f"field '{name}'", value, field.type)

    def _delete(self) -> None:
        """Mark record as deleted and remove it's ID from index."""
        if not self._state & STATE_DELETE:
            self._state |= STATE_DELETE
            self._delete_hook(self._id)

    def _update(self, **kwds: Any) -> None:
        """Mark record as updated and write the update to diff table."""
        if not self._state & STATE_UPDATE:
            self._state |= STATE_UPDATE
            self._update_hook(self._id, **kwds)

    def _restore(self) -> None:
        """Mark record as not deleted and append it's ID to index."""
        if self._state & STATE_DELETE:
            self._state &= ~STATE_DELETE
            self._restore_hook(self._id)

    def _revoke(self) -> None:
        """Mark record as not updated and remove the update from diff table."""
        if self._state & STATE_UPDATE:
            self._state &= ~STATE_UPDATE
            self._revoke_hook(self._id)

    @abc.abstractmethod
    def _get_newid(self) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def _delete_hook(self, rowid: int) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def _restore_hook(self, rowid: int) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def _update_hook(self, rowid: int, **kwds: Any) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def _revoke_hook(self, rowid: int) -> None:
        raise NotImplementedError()

#
# Builder, Constructor and helper functions
#

def build(*args: ColDef, newid: OptOp = None, **kwds: Any) -> Type[Record]:
    """Create a new subclass of the Record class.

    Args:
        *args: *column definitions*. All column definitions independent from
            each other can be given in one of the following formats: (1) In
            order to only specify the name of the column, without further
            information, the colum definition has to be given as a string
            `<name>`. Thereby the choice of `<name>` is restricted to valid
            identifiers, described by [UAX31]_. (2) If, additionally to the
            name, also the data type of the column shall be specified, then the
            column definition has to be given as a tuple `(<name>, <type>)`.
            Thereby `<type>` is required to by a valid :class:`type`, like like
            :class:`str`, :class:`int`, :class:`float` or :class:`Date
            <datetime.datetime>`. (3) Finally the column definition may also
            contain supplementary constraints and metadata. In this case the
            definition has to be given as a tuple `(<name>, <type>, <dict>)`,
            where `<dict>` is dictionary which comprises any items, documented
            by the function :func:`dataclasses.fields`.
        newid: Optional reference to a method, which returns the current ID of
            a new instance of the Record class. By default the Record class
            uses an internal Iterator.
        **kwds: Optional references to methods, which are bound to specific
            events of the new Record class. These events are: 'delete',
            'restore', 'update' and 'revoke'. By default no events are hooked.

    Returns:
        New subclass of the Record class.

    """
    # Check column defnitions and convert them to field descriptors, as required
    # by dataclasses.make_dataclass()
    fields: list = []
    names: StrList = []
    for arg in args:
        if isinstance(arg, str):
            fields.append(arg)
            names.append(arg)
            continue
        check.has_type(f'column {arg}', arg, tuple)
        check.has_size(f'column {arg}', arg, min_size=2, max_size=3)
        check.has_type('first argument', arg[0], str)
        check.has_type('second argument', arg[1], TypeHint)
        if len(arg) == 2:
            fields.append(arg)
            names.append(arg[0])
            continue
        check.has_type('third argument', arg[2], (Field, dict))
        if isinstance(arg[2], Field):
            fields.append(arg)
            names.append(arg[0])
            continue
        field = dataclasses.field(**arg[2])
        names.append(arg[0])
        fields.append(arg[:2] + (field,))

    # Dynamically create a dataclass, which is inherited from Record class.
    # Thereby create an ampty '__slots__' attribute to avoid collision with
    # default values (which in dataclasses are stored as class variables),
    # while avoiding the creation of a '__dict__' attribute
    namespace: StrDict = {}
    if newid and callable(newid):
        namespace['_get_newid'] = newid
    else:
        counter = itertools.count() # Infinite iterator
        namespace['_get_newid'] = lambda obj: next(counter)
    hooks = {
        'delete': '_delete_hook', 'restore': '_restore_hook',
        'update': '_update_hook', 'revoke': '_revoke_hook'}
    for key in hooks:
        namespace[hooks[key]] = kwds.get(key, void)
    namespace['__slots__'] = tuple()
    dataclass = dataclasses.make_dataclass(
        Record.__name__, fields, bases=(Record, ), namespace=namespace)

    # Dynamically create a new class, which is inherited from dataclass,
    # with corrected __slots__ attribute.
    return type(dataclass.__name__, (dataclass,), {'__slots__': names})

def create_from(record: Record, **changes: Any) -> Record:
    """Create new record instance, using an existing record."""
    if not is_record(record):
        raise InvalidTypeError('record', record, Record)
    newrec = dataclasses.replace(record, **changes)
    with contextlib.suppress(AttributeError):
        newrec._id = record._id # pylint: disable=W0212
    with contextlib.suppress(AttributeError):
        newrec._state = record._state # pylint: disable=W0212
    with contextlib.suppress(AttributeError):
        newrec._name = record._name # pylint: disable=W0212
    return newrec

def is_record(record: object) -> bool:
    """Check if a given object is a record."""
    if not isinstance(record, Record):
        return False
    if not dataclasses.is_dataclass(record):
        return False
    return True

def get_fields(record: Record) -> FieldTuple:
    """Get field definitions of record."""
    return dataclasses.fields(record)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Table Proxy for textfiles containing delimiter-separated values."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

from typing import Any
from flib.base import attrib
from flib.errors import ConnectError, DisconnectError
from flib.io import csv, ini
from flib.io import FileRef, FileRefClassInfo
from pandb.core import table

#
# CSV Table Proxy Class
#

class Table(table.Proxy):
    """CSV-Table Proxy."""

    #
    # Public Attributes
    #

    # Remove setter from name, since the name is determined by the file name
    name: property = attrib.Virtual('_get_name')
    name.__doc__ = "Name of the table."

    #
    # Protected Attributes
    #

    _fileref: property = attrib.Temporary(dtype=FileRefClassInfo)
    _file: property = attrib.Temporary(dtype=csv.File)

    #
    # Special Methods
    #

    def __init__(self, file: FileRef, *args: Any, **kwds: Any) -> None:
        """Initialize CSV-Table Proxy.

        Args:
            file:
            *args: Additional arguments, that are passed to
                :class:`csv.File <flib.base.io.csv.File>`.
            **kwds: Additional keyword arguments, that are passed to csv.File.

        """
        super().__init__() # Initialize table proxy
        self.connect(file, *args, **kwds) # Connect CSV File
        self._post_init() # Run post init hook

    #
    # Public Methods
    #

    def connect( # type: ignore
            self, file: FileRef, *args: Any, **kwds: Any) -> None:
        """Connect to given file reference."""
        if self._connected:
            raise ConnectError("the connection already has been established")

        # Create and reference CSV file
        fh = csv.File(file, *args, **kwds)
        self._file = fh

        # Get table name, structure and and metadata from CSV file
        name = fh.name.split('.', 1)[0]
        fields = fh.fields
        metadata = ini.decode(fh.comment, flat=True, autocast=True)

        # Create table
        self.create(name, fields, metadata=metadata)
        self._connected = True

    def disconnect(self) -> None:
        """Close connection to referenced file."""
        if not self._connected:
            raise DisconnectError("the proxy has not yet been connected")
        self._file.close()
        self._connected = False

    def pull(self) -> None:
        """Pull all rows from CSV-File."""
        comment = self._file.comment
        mapping = ini.decode(comment, flat=True, autocast=True)
        self._metadata.update(mapping)
        name = self._file.name.split('.', 1)[0] # Get name from filename
        self._set_name(name)
        rows = self._file.read()
        self.insert(rows)

    def push(self) -> None:
        """Push all rows to CSV-File."""
        comment = ini.encode(self.metadata, flat=True)
        self._file.comment = comment
        rows = self.select()
        self._file.write(rows)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Data Types.

This module comprises a collection of data types, as specified in the SQL
standard. Thereby in most cases the types are directly taken from Python3
builtin or standard library types. For fixed length arrays the corresponding
ctypes type is used.

"""
__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

#import ctypes
import decimal
import datetime
import uuid

#
# Numeric Types
#

# Problem: Operators do not work!
# SmallInt = ctypes.c_short
# Integer = ctypes.c_int
# BigInt = ctypes.c_longlong
# Real = ctypes.c_float
Decimal = decimal.Decimal
Integer = int
Float = float

#
# Character Types
#

Varchar = str
Char = str # ctypes.create_string_buffer

#
# Binary Types
#

#
# Large object types
#


#
# Misc Types
#

Boolean = bool
UUID = uuid.UUID

#
# Date and time Types
#

Date = datetime.date
Time = datetime.time
TimeStamp = datetime.datetime



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""DB-API 2.0 database interface scaffold.

This module is a reference for the required exceptions, base classes, module
attributes and module functions specified in the Python Database API (DB-API)
Specification 2.0 :PEP:`249`.

Module attributes:
    apilevel:
        String constant stating the supported DB-API level.
    threadsafety:
        Integer constant stating the level of thread safety the interface
        supports.
    paramstyle:
        String constant stating the type of parameter marker formatting expected
        by the interface.

Constructor function:
    connect(...) -> Connection
        Constructor for creating a connection to the database. Returns a
        Connection Object. It takes a number of parameters which are database
        dependent.

"""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

from typing import Any
from abc import ABC, abstractmethod
from flib.base import attrib, errors
from flib.base.types import OptList, OptInt, OptBool

#
# DB-API 2.0 Exceptions
#

class Error(errors.UserError):
    """DB-API Error.

    Exception that is the base class of all other error exceptions. You can use
    this to catch all errors with one single except statement. Warnings are not
    considered errors and thus should not use this class as base. It must be a
    subclass of the Python StandardError (defined in the module exceptions).
    """

class Warning(errors.UserError): # pylint: disable=W0622
    """DB-API Warning.

    Exception raised for important warnings like data truncations while
    inserting, etc. It must be a subclass of the Standard DBIError.
    """

class InterfaceError(Error):
    """DB-API InterfaceError.

    Exception raised for errors that are related to the database interface
    rather than the database itself. It must be a subclass of Error.
    """

class DatabaseError(Error):
    """DB-API DatabaseError.

    Exception raised for errors that are related to the database. It must be a
    subclass of Error.
    """

class InternalError(DatabaseError):
    """DB-API InternalError.

    Exception raised when the database encounters an internal error, e.g. the
    cursor is not valid anymore, the transaction is out of sync, etc. It must
    be a subclass of DatabaseError.
    """

class OperationalError(DatabaseError):
    """DB-API OperationalError.

    Exception raised for errors that are related to the database's operation and
    not necessarily under the control of the programmer, e.g. an unexpected
    disconnect occurs, the data source name is not found, a transaction could
    not be processed, a memory allocation error occurred during processing, etc.
    It must be a subclass of DatabaseError.
    """

class ProgrammingError(DatabaseError):
    """DB-API ProgrammingError.

    Exception raised for programming errors, e.g. table not found or already
    exists, syntax error in the SQL statement, wrong number of parameters
    specified, etc. It must be a subclass of DatabaseError.
    """

class IntegrityError(DatabaseError):
    """DB-API IntegrityError.

    Exception raised when the relational integrity of the database is affected,
    e.g. a foreign key check fails. It must be a subclass of DatabaseError.
    """

class DataError(DatabaseError):
    """DB-API DataError.

    Exception raised for errors that are due to problems with the processed data
    like division by zero, numeric value out of range, etc. It must be a
    subclass of DatabaseError.
    """

class NotSupportedError(DatabaseError):
    """DB-API NotSupportedError.

    Exception raised in case a method or database API was used which is not
    supported by the database, e.g. requesting a `rollback` on a connection
    that does not support transaction or has transactions turned off. It must be
    a subclass of DatabaseError.
    """

#
# DB-API 2.0 Cursor Class
#

class Cursor(attrib.Group, ABC):
    """Database Cursor.

    These objects represent a database cursor, which is used to manage the
    context of a fetch operation. Cursors created from the same connection are
    not isolated, i.e., any changes done to the database by a cursor are
    immediately visible by the other cursors. Cursors created from different
    connections can or can not be isolated, depending on how the transaction
    support is implemented (see also the connection's .rollback() and .commit()
    methods).

    """

    #
    # Cursor attributes
    #

    arraysize: property = attrib.MetaData(dtype=int, default=1)
    arraysize.__doc__ = """
    This read/write attribute specifies the number of rows to fetch at a time
    with `fetchmany`. It defaults to 1 meaning to fetch a single row at a time.
    Implementations must observe this value with respect to the `fetchmany`
    method, but are free to interact with the database a single row at a time.
    It may also be used in the implementation of `executemany`.
    """

    description: property = attrib.Virtual('_get_description')
    description.__doc__ = """
    Sequence of 7-item sequences containing information about one result column:
    name, type_code, display_size, internal_size, precision, scale, null_ok
    The first two items (name and type_code) are mandatory, the other five are
    optional and are set to None if no meaningful values can be provided.
    This attribute will be None for operations that do not return rows or if the
    cursor has not had an operation invoked via the .execute*() method yet.
    """

    rowcount: property = attrib.Virtual('_get_rowcount')
    description.__doc__ = """
    This read-only attribute specifies the number of rows that the last
    execute*() produced (for DQL statements like SELECT) or affected (for DML
    statements like UPDATE or INSERT). The attribute is -1 in case no
    .execute*() has been performed on the cursor or the rowcount of the last
    operation is cannot be determined by the interface.
    """

    #
    # Public Methods
    #

    @abstractmethod
    def callproc(self, procname: str, *args: Any, **kwds: Any) -> Any:
        """Call stored database procedure.

        Call a stored database procedure with the given name. The sequence of
        parameters must contain one entry for each argument, that the procedure
        expects. The result of the call is returned as modified copy of the
        input sequence. Input parameters are left untouched, output and
        input/output parameters replaced with possibly new values.

        The procedure may also provide a result set as output. This must then be
        made available through the standard .fetch*() methods.

        If the database does not support the functionality required by the
        method, the interface should throw an exception in case the method is
        used.
        """

    @abstractmethod
    def close(self) -> None:
        """Close the cursor now (rather than whenever __del__ is called).

        The cursor will be unusable from this point forward; an Error (or
        subclass) exception will be raised if any operation is attempted with
        the cursor.
        """

    @abstractmethod
    def execute(self, operation: str, *args: Any) -> Any:
        """Prepare and execute a database operation (query or command).

        Parameters may be provided as sequence or mapping and will be bound to
        variables in the operation. Variables are specified in a
        database-specific notation, which is identified by the module global
        `paramstyle`.

        A reference to the operation will be retained by the cursor. If the same
        operation object is passed in again, then the cursor can optimize its
        behavior. This is most effective for algorithms where the same operation
        is used, but different parameters are bound to it.

        For maximum efficiency when reusing an operation, it is best to use the
        `setinputsizes` method to specify the parameter types and sizes ahead
        of time. It is legal for a parameter to not match the predefined
        information; the implementation should compensate, possibly with a loss
        of efficiency.

        The parameters may also be specified as list of tuples to e.g. insert
        multiple rows in a single operation, but this kind of usage is
        deprecated: .executemany() should be used instead.
        """

    @abstractmethod
    def executemany(self, operation: str, seq_of_parameters: list) -> Any:
        """Prepare and execute database operation for multiple parameters.

        Prepare a database operation (query or command) and then execute it
        against all parameter sequences or mappings found in the sequence
        *seq_of_parameters*.

        Modules are free to implement this method using multiple calls to the
        `execute` method or by using array operations to have the database
        process the sequence as a whole in one call.

        Use of this method for an operation which produces one or more result
        sets constitutes undefined behavior, and the implementation is permitted
        (but not required) to raise an exception when it detects that a result
        set has been created by an invocation of the operation.
        """

    @abstractmethod
    def fetchone(self) -> OptList:
        """Fetch the next row of a query result.

        Fetch the next row of a query result set, returning a single sequence,
        or None when no more data is available.

        An Error (or subclass) exception is raised if the previous call to
        `execute` did not produce any result set or no call was issued yet.
        """


    @abstractmethod
    def fetchmany(self, size: OptInt) -> list:
        """Fetch the next set of rows of a query result.

        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when no
        more rows are available.

        The number of rows to fetch per call is specified by the parameter. If
        it is not given, the cursor's arraysize determines the number of rows to
        be fetched. The method should try to fetch as many rows as indicated by
        the size parameter. If this is not possible due to the specified number
        of rows not being available, fewer rows may be returned.

        An Error (or subclass) exception is raised if the previous call to
        `execute` did not produce any result set or no call was issued yet.

        Note there are performance considerations involved with the size
        parameter. For optimal performance, it is usually best to use the
        `arraysize` attribute. If the size parameter is used, then it is best
        for it to retain the same value from one `fetchmany` call to the next.
        """

    @abstractmethod
    def fetchall(self) -> list:
        """Fetch all remaining rows of a query result.

        Fetch all remaining rows of a query result, returning them as a sequence
        of sequences (e.g. a list of tuples). Note that the cursor's arraysize
        attribute can affect the performance of this operation.

        An Error (or subclass) exception is raised if the previous call to
        `execute` did not produce any result set or no call was issued yet.
        """

    @abstractmethod
    def nextset(self) -> OptBool:
        """Skip cursor to the next available set (if supported).

        This method will make the cursor skip to the next available set,
        discarding any remaining rows from the current set. If there are no more
        sets, the method returns None. Otherwise, it returns a true value and
        subsequent calls to the `fetch*` methods will return rows from the next
        result set.

        An Error (or subclass) exception is raised if the previous call to an
        `execute*` method did not produce any result set or no call was issued
        yet.

        If the database does not support the functionality required by the
        method, the interface should throw an exception in case the method is
        used.
        """

    @abstractmethod
    def setinputsizes(self, sizes: list) -> None:
        """Set input sizes for database operations (query or command).

        This can be used before a call to `execute*` to predefine memory areas
        for the operation's parameters. *sizes* is specified as a sequence with
        one item for each input parameter. The item should be a type object that
        corresponds to the input that will be used, or it should be an integer
        specifying the maximum length of a string parameter. If the item is
        None, then no predefined memory area will be reserved for that column
        (this is useful to avoid predefined areas for large inputs).

        This method would be used before the `execute*` method is invoked.
        Implementations are free to have this method do nothing and users are
        free to not use it.
        """

    @abstractmethod
    def setoutputsize(self, size: int, column: OptInt) -> None:
        """Set a column buffer size for fetches of large columns.

        The column is specified as an index into the result sequence. Not
        specifying the column will set the default size for all large columns in
        the cursor.

        This method would be used before an `execute*` method is invoked.
        Implementations are free to have this method do nothing and users are
        free to not use it.
        """

    #
    # Protected Methods
    #

    @abstractmethod
    def _get_description(self) -> list:
        pass

    @abstractmethod
    def _get_rowcount(self) -> int:
        pass

#
# DB-API 2.0 Connection Class
#

class Connection(attrib.Group, ABC):
    """Database Connection."""

    @abstractmethod
    def close(self) -> None:
        """Close the connection now.

        The connection will be unusable from this point forward; an Error (or
        subclass) exception will be raised if any operation is attempted with
        the connection. The same applies to all cursor objects trying to use the
        connection. Note that closing a connection without committing the
        changes first will cause an implicit rollback to be performed.
        """

    @abstractmethod
    def commit(self) -> None:
        """Commit any pending transaction to the database.

        Note that if the database supports an auto-commit feature, this must be
        initially off. An interface method may be provided to turn it back on.
        Database modules that do not support transactions should implement this
        method with void functionality.
        """

    @abstractmethod
    def rollback(self) -> None:
        """Rollback transaction (if supported).

        In case a database does provide transactions this method causes the
        database to roll back to the start of any pending transaction. Closing a
        connection without committing the changes first will cause an implicit
        rollback to be performed.
        """

    @abstractmethod
    def cursor(self) -> Cursor:
        """Return a new Cursor Object using the connection.

        If the database does not provide a direct cursor concept, the module
        will have to emulate cursors using other means to the extent needed by
        this specification.
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Frootlab Developers
#
# This file is part of Pandora, https://frootlab.github.io/pandora
#
#  Pandora is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Pandora is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  Pandora. If not, see <http://www.gnu.org/licenses/>.
#
"""Setuptools based installation."""

__author__ = 'Patrick Michl'
__email__ = 'frootlab@gmail.com'
__license__ = 'GPLv3'
__docformat__ = 'google'

import pathlib
import re
import setuptools

def install() -> None:
    """Setuptools based installation script."""

    #Get module variables from file 'pandb/__init__.py'
    text = pathlib.Path('./pandb/__init__.py').read_text()
    rekey = "__([a-zA-Z][a-zA-Z0-9_]*)__"
    reval = r"['\"]([^'\"]*)['\"]"
    pattern = f"^[ ]*{rekey}[ ]*=[ ]*{reval}"
    pkg_vars = {}
    for mo in re.finditer(pattern, text, re.M):
        pkg_vars[str(mo.group(1))] = str(mo.group(2))

    # Install package
    setuptools.setup(
        name='pandb',
        version=pkg_vars['version'],
        description='pandora database proxy',
        long_description=pathlib.Path('.', 'README.rst').read_text(),
        long_description_content_type='text/x-rst',
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Programming Language :: Python :: 3',
    		'Programming Language :: Python :: 3.7',
            'Operating System :: OS Independent',
            'Topic :: Database :: Database Engines/Servers',
            'Topic :: Software Development :: Libraries :: Python Modules'],
        keywords=(
            'database '
            'database-proxy '
            'orm-framework '
            'data-warehouse '),
        url='https://frootlab.github.io/pandora',
        author=pkg_vars['author'],
        author_email=pkg_vars['email'],
        license=pkg_vars['license'],
        packages=setuptools.find_packages(exclude=['docs', 'tests']),
        package_dir={
            'pandb': 'pandb'},
        python_requires='>=3.7',
        install_requires=[
            'numpy>=1.15',
            'SQLAlchemy>=1.2.17',
            'flib-errors>=0.0.4',
            'flib-base>=0.1.18',
            'flib-io>=0.1.6']
    )

if __name__ == '__main__':
    install()
